import asyncio
from concurrent.futures import Executor
from multiprocessing import Pool, cpu_count
from typing import Any, Callable, Set, Tuple, TypeVar

from .compat import EventLoopMixin
from .counters import Statistic
from .utils import set_exception


T = TypeVar("T")
FuturesSet = Set[asyncio.Future]
_CreateFutureType = Tuple[
    Callable[[T], None],
    Callable[[T], None],
    asyncio.Future,
]


class ProcessPoolStatistic(Statistic):
    processes: int
    done: int
    error: int
    success: int
    submitted: int
    sum_time: float


class ProcessPoolExecutor(Executor, EventLoopMixin):
    __slots__ = (
        "__futures", "__pool", "_statistic",
    ) + EventLoopMixin.__slots__

    DEFAULT_MAX_WORKERS = max((cpu_count(), 4))

    def __init__(self, max_workers: int = DEFAULT_MAX_WORKERS, **kwargs: Any):
        self.__futures: FuturesSet = set()
        self.__pool = Pool(processes=max_workers, **kwargs)
        self._statistic = ProcessPoolStatistic()
        self._statistic.processes = max_workers

    def _create_future(self) -> _CreateFutureType:
        future = self.loop.create_future()  # type: asyncio.Future

        self.__futures.add(future)
        future.add_done_callback(self.__futures.discard)

        start_time = self.loop.time()

        def callback(result: T) -> None:
            self._statistic.success += 1
            self._statistic.done += 1
            self._statistic.sum_time += self.loop.time() - start_time
            self.loop.call_soon_threadsafe(future.set_result, result)

        def errorback(exc: T) -> None:
            self._statistic.error += 1
            self._statistic.done += 1
            self._statistic.sum_time += self.loop.time() - start_time
            self.loop.call_soon_threadsafe(future.set_exception, exc)

        return callback, errorback, future

    def submit(           # type: ignore
        self, fn: Callable[..., T],
        *args: Any, **kwargs: Any,
    ) -> asyncio.Future:
        """
        Submit blocking function to the pool
        """
        if fn is None or not callable(fn):
            raise ValueError("First argument must be callable")

        callback, errorback, future = self._create_future()

        self.__pool.apply_async(
            fn,
            args=args,
            kwds=kwargs,
            callback=callback,
            error_callback=errorback,
        )

        self._statistic.submitted += 1
        return future

    # noinspection PyMethodOverriding
    def shutdown(
        self, wait: bool = True, cancel_futures: bool = False,
    ) -> None:

        if not self.__pool:
            return

        self.__pool.terminate()

        if cancel_futures:
            set_exception(self.__futures, asyncio.CancelledError())

        self.__pool.close()

        if wait:
            self.__pool.join()

    def __del__(self) -> None:
        self.shutdown()



import asyncio
import hashlib
import hmac
import os
import socket
import sys
from inspect import Traceback
from multiprocessing import ProcessError
from os import chmod, urandom
from subprocess import PIPE, Popen
from tempfile import mktemp
from types import MappingProxyType
from typing import (
    Any, Callable, Coroutine, Dict, Mapping, Optional, Set, Tuple, Type,
)

from aiomisc.counters import Statistic
from aiomisc.thread_pool import threaded
from aiomisc.utils import (
    bind_socket, cancel_tasks, fast_uuid4, set_exception, shield,
)
from aiomisc_log import LOG_FORMAT, LOG_LEVEL
from aiomisc_worker import (
    COOKIE_SIZE, INET_AF, INT_SIGNAL, SIGNAL, AddressType, PacketTypes, T, log,
)
from aiomisc_worker.protocol import AsyncProtocol, FileIOProtocol


class WorkerPoolStatistic(Statistic):
    processes: int
    spawning: int
    queue_size: int
    submitted: int
    sum_time: float
    done: int
    success: int
    error: int
    bad_auth: int
    task_added: int


class WorkerPool:
    tasks: asyncio.Queue
    server: asyncio.AbstractServer
    address: AddressType
    initializer: Optional[Callable[[], Any]]
    initializer_args: Tuple[Any, ...]
    initializer_kwargs: Mapping[str, Any]

    _supervisor: Popen
    worker_ids: Tuple[bytes, ...]
    pids: Set[int]

    if hasattr(socket, "AF_UNIX"):
        def _create_socket(self) -> None:
            path = mktemp(suffix=".sock", prefix="worker-")
            self.socket = bind_socket(
                socket.AF_UNIX,
                socket.SOCK_STREAM,
                address=path,
            )
            self.address = path
            chmod(path, 0o600)
    else:
        def _create_socket(self) -> None:
            self.socket = bind_socket(
                INET_AF,
                socket.SOCK_STREAM,
                address="localhost",
                reuse_addr=False,
                reuse_port=False,
            )
            self.address = self.socket.getsockname()[:2]

    @staticmethod
    def _kill_process(process: Popen) -> None:
        if process.returncode is not None:
            return None
        log.debug("Terminating worker pool process PID: %s", process.pid)
        process.kill()

    @threaded
    def __create_supervisor(self, *identity: str) -> Popen:
        if self.__closing:
            raise RuntimeError("Pool closed")

        env = dict(os.environ)
        env["AIOMISC_NO_PLUGINS"] = ""
        process = Popen(
            [sys.executable, "-m", "aiomisc_worker"], stdin=PIPE, env=env,
        )

        assert process.stdin

        log_level = (
            log.getEffectiveLevel()
            if LOG_LEVEL is None
            else LOG_LEVEL.get()
        )

        log_format = "color" if LOG_FORMAT is None else LOG_FORMAT.get()

        proto_stdin = FileIOProtocol(process.stdin)

        proto_stdin.send((log_level, log_format))
        proto_stdin.send(self.address)
        proto_stdin.send(self.__cookie)
        proto_stdin.send(identity)
        proto_stdin.send((
            self.initializer,
            self.initializer_args,
            dict(self.initializer_kwargs),
        ))
        process.stdin.close()
        return process

    def __init__(
        self, workers: int, max_overflow: int = 0, *,
        initializer: Optional[Callable[..., Any]] = None,
        initializer_args: Tuple[Any, ...] = (),
        initializer_kwargs: Mapping[str, Any] = MappingProxyType({}),
        statistic_name: Optional[str] = None,
    ):
        self._create_socket()
        self.__cookie = urandom(COOKIE_SIZE)
        self.__loop: Optional[asyncio.AbstractEventLoop] = None
        self.__futures: Set[asyncio.Future] = set()
        self.__task_store: Set[asyncio.Task] = set()
        self.__closing = False
        self.__closing_lock = asyncio.Lock()
        self._statistic = WorkerPoolStatistic(name=statistic_name)
        self.__max_overflow = max_overflow
        self.workers = workers
        self.pids = set()
        self.initializer = initializer
        self.initializer_args = initializer_args
        self.initializer_kwargs = initializer_kwargs

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        if self.__loop is None:
            self.__loop = asyncio.get_running_loop()
        return self.__loop

    async def __handle_client(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter,
    ) -> None:
        proto = AsyncProtocol(reader, writer)
        packet_type, worker_id, digest, pid = await proto.receive()

        async with self.__closing_lock:
            if self.__closing:
                proto.close()

        if packet_type == PacketTypes.BAD_INITIALIZER:
            packet_type, exc = await proto.receive()
            if packet_type != PacketTypes.EXCEPTION:
                await proto.send(PacketTypes.BAD_PACKET)
            else:
                set_exception(self.__futures, exc)
            await self.close()
            return

        if packet_type != PacketTypes.AUTH:
            await proto.send(PacketTypes.BAD_PACKET)
            if writer.can_write_eof():
                writer.write_eof()
            return

        if worker_id not in self.worker_ids:
            log.error("Unknown worker with id %r", worker_id)
            return

        expected_digest = hmac.HMAC(
            self.__cookie,
            worker_id,
            digestmod=hashlib.sha256,
        ).digest()

        if expected_digest != digest:
            await proto.send(PacketTypes.AUTH_FAIL)
            if writer.can_write_eof():
                writer.write_eof()
            log.debug("Bad digest %r expected %r", digest, expected_digest)
            return

        await proto.send(PacketTypes.AUTH_OK)
        self._statistic.processes += 1
        self._statistic.spawning += 1
        self.pids.add(pid)

        try:
            while not reader.at_eof():
                func: Callable
                args: Tuple[Any, ...]
                kwargs: Dict[str, Any]
                result_future: asyncio.Future
                process_future: asyncio.Future

                (
                    func, args, kwargs, result_future, process_future,
                ) = await self.tasks.get()

                if process_future.done() or result_future.done():
                    continue

                try:
                    process_future.set_result(pid)
                    await proto.send((PacketTypes.REQUEST, func, args, kwargs))
                    packet_type, payload = await proto.receive()

                    if result_future.done():
                        log.debug(
                            "Result future %r already done, skipping",
                            result_future,
                        )
                        continue

                    if packet_type == PacketTypes.RESULT:
                        result_future.set_result(payload)
                    elif packet_type in (
                        PacketTypes.EXCEPTION, PacketTypes.CANCELLED,
                    ):
                        result_future.set_exception(payload)
                    del packet_type, payload
                except (asyncio.IncompleteReadError, ConnectionError):
                    if not result_future.done():
                        result_future.set_exception(
                            ProcessError(f"Process {pid!r} unexpected exited"),
                        )
                    break
                except Exception as e:
                    if not result_future.done():
                        result_future.set_exception(e)

                    if not writer.is_closing():
                        if writer.can_write_eof():
                            writer.write_eof()
                        writer.close()
                    raise
        finally:
            self._statistic.processes -= 1
            self.pids.remove(pid)

    def __start_handler(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter,
    ) -> asyncio.Task:
        return self.__task(self.__handle_client(reader, writer))

    def __task_add(self, task: asyncio.Task) -> None:
        self._statistic.task_added += 1
        task.add_done_callback(self.__task_store.remove)
        self.__task_store.add(task)

    def __task(self, coroutine: Coroutine) -> asyncio.Task:
        task = self.loop.create_task(coroutine)
        self.__task_add(task)
        return task

    async def start(self) -> None:
        self.tasks = asyncio.Queue(maxsize=self.__max_overflow)
        self.server = await asyncio.start_server(
            self.__start_handler,
            sock=self.socket,
        )
        del self.socket

        self.worker_ids = tuple(
            fast_uuid4().bytes for _ in range(self.workers)
        )
        self._supervisor = await self.__create_supervisor(*self.worker_ids)

    def __create_future(self) -> asyncio.Future:
        future = self.loop.create_future()
        self.__futures.add(future)
        future.add_done_callback(self.__futures.remove)
        return future

    def __reject_futures(self) -> None:
        set_exception(self.__futures, RuntimeError("Pool closed"))

    @shield
    async def close(self) -> None:
        async with self.__closing_lock:
            if self.__closing:
                return

            self._kill_supervisor()
            self.__closing = True
            self.server.close()
            await self.server.wait_closed()

            await cancel_tasks(tuple(self.__task_store))
            await cancel_tasks(tuple(self.__futures))

    def _kill_supervisor(self) -> None:
        supervisor: Optional[Popen] = getattr(self, "_supervisor", None)
        if supervisor is None or supervisor.poll() is not None:
            return

        log.debug(
            "Sending %r to supervisor process PID: %d. Workers: %r",
            INT_SIGNAL, supervisor.pid, self.pids,
        )
        os.kill(self._supervisor.pid, INT_SIGNAL)

    def __del__(self) -> None:
        self._kill_supervisor()

    async def create_task(
        self, func: Callable[..., T],
        *args: Any, **kwargs: Any,
    ) -> T:
        result_future = self.__create_future()
        process_future = self.__create_future()

        await self.tasks.put((
            func, args, kwargs, result_future, process_future,
        ))

        pid: int = await process_future

        try:
            return await result_future
        except asyncio.CancelledError:
            log.debug("Sending %r to worker PID: %d", SIGNAL, pid)
            os.kill(pid, SIGNAL)
            raise

    async def __aenter__(self) -> "WorkerPool":
        await self.start()
        return self

    async def __aexit__(
        self, exc_type: Type[Exception],
        exc_val: Exception, exc_tb: Traceback,
    ) -> None:
        await self.close()



import abc
import asyncio
import logging
import socket
from http import HTTPStatus
from types import MappingProxyType
from typing import Any, Awaitable, Callable, Iterable, Mapping, Optional, Set

import aiohttp
import yarl
from aiohttp import ClientSession, TCPConnector
from raven import Client  # type: ignore
from raven.conf import defaults  # type: ignore
from raven.exceptions import APIError, RateLimited  # type: ignore
from raven.handlers.logging import SentryHandler  # type: ignore
from raven.transport import Transport  # type: ignore
from raven.transport.base import AsyncTransport  # type: ignore
from raven.transport.http import HTTPTransport  # type: ignore

from aiomisc.service import Service
from aiomisc.utils import TimeoutType


log = logging.getLogger(__name__)


__doc__ = """
This module based on https://github.com/getsentry/raven-aiohttp

Copyright (c) 2018 Functional Software, Inc and individual contributors.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

    3. Neither the name of the Raven nor the names of its contributors may be
    used to endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
"""


class DummyTransport(Transport):         # type: ignore
    def send(self, url, data, headers):  # type: ignore
        pass


class AioHttpTransportBase(
    AsyncTransport, HTTPTransport, metaclass=abc.ABCMeta,    # type: ignore
):

    def __init__(
        self, parsed_url: Optional[str] = None, *, verify_ssl: bool = True,
        timeout: TimeoutType = defaults.TIMEOUT, keepalive: bool = True,
        family: int = socket.AF_INET,
    ):
        self._keepalive = keepalive
        self._family = family
        self._loop = asyncio.get_event_loop()

        if parsed_url is not None:
            raise TypeError(
                "Transport accepts no URLs for this version of raven.",
            )
        super().__init__(timeout, verify_ssl)

        if self.keepalive:
            self._client_session = self._client_session_factory()

        self._closing = False

    @property
    def keepalive(self) -> bool:
        return self._keepalive

    @property
    def family(self) -> int:
        return self._family

    def _client_session_factory(self) -> aiohttp.ClientSession:
        connector = aiohttp.TCPConnector(
            verify_ssl=self.verify_ssl, family=self.family,
        )
        return aiohttp.ClientSession(
            connector=connector,
        )

    async def _do_send(
        self, url: str, data: Any, headers: Mapping,
        callback: Callable[[], Any], errorback: Callable[[Any], Any],
    ) -> None:
        if self.keepalive:
            session = self._client_session
        else:
            session = self._client_session_factory()

        resp = None

        try:
            resp = await session.post(
                url, data=data, compress=False,
                headers=headers, timeout=self.timeout,
            )

            code = resp.status
            if code != HTTPStatus.OK:
                msg = resp.headers.get("x-sentry-error")
                if code == HTTPStatus.TOO_MANY_REQUESTS:
                    try:
                        retry_after = int(resp.headers.get("retry-after", "0"))
                    except (ValueError, TypeError):
                        retry_after = 0
                    errorback(RateLimited(msg, retry_after))
                else:
                    errorback(APIError(msg, code))
            else:
                callback()
        except asyncio.CancelledError:
            # do not mute asyncio.CancelledError
            raise
        except Exception as exc:
            errorback(exc)
        finally:
            if resp is not None:
                resp.release()
            if not self.keepalive:
                await session.close()

    @abc.abstractmethod
    def _async_send(
        self, url: str, data: Any, headers: Mapping,
        success_cb: Callable[[], Any], failure_cb: Callable[[Any], Any],
    ) -> None:  # pragma: no cover
        pass

    @abc.abstractmethod
    async def _close(self) -> None:  # pragma: no cover
        pass

    def async_send(
        self, url: str, data: Any, headers: Mapping,
        success_cb: Callable[[], Any], failure_cb: Callable[[Any], Any],
    ) -> None:
        if self._closing:
            failure_cb(RuntimeError(f"{self.__class__.__name__} is closed"))
            return

        self._loop.call_soon_threadsafe(
            self._async_send, url, data, headers, success_cb, failure_cb,
        )

    async def _close_coro(
        self,
        *,
        timeout: Optional[TimeoutType] = None,
    ) -> None:
        try:
            await asyncio.wait_for(
                self._close(), timeout=timeout,
            )
        except asyncio.TimeoutError:
            pass
        finally:
            if self.keepalive:
                await self._client_session.close()

    def close(self, *, timeout: Optional[TimeoutType] = None) -> Awaitable[Any]:
        if self._closing:
            async def dummy() -> None:
                pass

            return dummy()

        self._closing = True

        return self._loop.create_task(self._close_coro(timeout=timeout))


class AioHttpTransport(AioHttpTransportBase):

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

        self._tasks: Set[asyncio.Task] = set()

    def _async_send(
        self, url: str, data: Any, headers: Mapping,
        success_cb: Callable[[], Any], failure_cb: Callable[[Any], Any],
    ) -> None:
        coro = self._do_send(url, data, headers, success_cb, failure_cb)

        task = self._loop.create_task(coro)
        self._tasks.add(task)
        task.add_done_callback(self._tasks.remove)

    async def _close(self) -> None:
        await asyncio.gather(
            *self._tasks,
            return_exceptions=True,
        )

        assert len(self._tasks) == 0


class QueuedAioHttpTransport(AioHttpTransportBase):

    def __init__(
        self, *args: Any, workers: int = 1, qsize: int = 1000, **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)

        self._queue: asyncio.Queue = asyncio.Queue(maxsize=qsize)

        self._workers = set()

        for _ in range(workers):
            worker: asyncio.Task = self._loop.create_task(self._worker())
            self._workers.add(worker)
            worker.add_done_callback(self._workers.remove)

    async def _worker(self) -> None:
        while True:
            data = await self._queue.get()

            try:
                if data is ...:
                    self._queue.put_nowait(...)
                    break

                url, data, headers, success_cb, failure_cb = data

                await self._do_send(
                    url, data, headers, success_cb, failure_cb,
                )
            finally:
                self._queue.task_done()

    def _async_send(
        self, url: str, data: Any, headers: Mapping,
        callback: Callable[[], Any], errorback: Callable[[Any], Any],
    ) -> None:
        payload = url, data, headers, callback, errorback

        try:
            self._queue.put_nowait(payload)
        except asyncio.QueueFull:
            skipped = self._queue.get_nowait()
            self._queue.task_done()

            *_, errorback = skipped

            errorback(
                RuntimeError("QueuedAioHttpTransport internal queue is full"),
            )

            self._queue.put_nowait(payload)

    async def _close(self) -> None:
        try:
            self._queue.put_nowait(...)
        except asyncio.QueueFull:
            skipped = self._queue.get_nowait()
            self._queue.task_done()

            *_, failure_cb = skipped

            failure_cb(
                RuntimeError("QueuedAioHttpTransport internal queue was full"),
            )

            self._queue.put_nowait(...)

        await asyncio.gather(
            *self._workers,
            return_exceptions=True,
        )

        assert len(self._workers) == 0
        assert self._queue.qsize() == 1
        try:
            assert self._queue.get_nowait() is ...
        finally:
            self._queue.task_done()


class QueuedKeepaliveAioHttpTransport(QueuedAioHttpTransport):
    DNS_CACHE_TTL = 600
    DNS_CACHE = True
    TCP_CONNECTION_LIMIT = 32
    TCP_CONNECTION_LIMIT_HOST = 8
    WORKERS = 1
    QUEUE_SIZE = 1000

    def __init__(
        self, *args: Any, family: int = socket.AF_UNSPEC,
        dns_cache: bool = DNS_CACHE,
        dns_cache_ttl: int = DNS_CACHE_TTL,
        connection_limit: int = TCP_CONNECTION_LIMIT,
        connection_limit_host: int = TCP_CONNECTION_LIMIT_HOST,
        workers: int = WORKERS, qsize: int = QUEUE_SIZE, **kwargs: Any,
    ):
        self.connection_limit = connection_limit
        self.connection_limit_host = connection_limit_host
        self.dns_cache = dns_cache
        self.dns_cache_ttl = dns_cache_ttl

        super().__init__(
            *args, family=family, keepalive=True,
            workers=workers, qsize=qsize, **kwargs,
        )

    def _client_session_factory(self) -> ClientSession:
        self.connector = TCPConnector(
            family=self.family,
            limit=self.connection_limit,
            limit_per_host=self.connection_limit_host,
            ttl_dns_cache=self.dns_cache_ttl,
            use_dns_cache=self.dns_cache,
            verify_ssl=self.verify_ssl,
        )

        return ClientSession(
            connector=self.connector,
            connector_owner=False,
        )

    async def _close(self) -> None:
        await super()._close()
        await self.connector.close()


class RavenSender(Service):
    __required__ = "sentry_dsn",

    sentry_dsn: yarl.URL
    min_level: int = logging.WARNING
    client_options: Mapping[str, Any] = MappingProxyType({})
    client: Client
    filters: Iterable[logging.Filter] = ()

    async def start(self) -> None:
        self.client = Client(
            str(self.sentry_dsn),
            transport=QueuedKeepaliveAioHttpTransport,
            **self.client_options,
        )

        # Initialize Transport object
        self.client.remote.get_transport()

        self.sentry_dsn = yarl.URL(
            self.sentry_dsn,
        ).with_password("").with_user("")

        log.info("Starting Raven for %r", self.sentry_dsn)

        handler = SentryHandler(client=self.client, level=self.min_level)

        # Add filters
        for fltr in self.filters:
            handler.addFilter(fltr)

        logging.getLogger().handlers.append(handler)

    async def stop(self, *_: Any) -> None:
        transport = self.client.remote.get_transport()
        self.client.set_dsn("", transport=DummyTransport)

        await transport.close()

        log.info("Stopping Raven client for %r", self.sentry_dsn)


__all__ = ("RavenSender",)



import logging
import os
import socket
from typing import Any, Iterable, Iterator, Optional, Tuple

from aiomisc.compat import get_current_loop
from aiomisc.periodic import PeriodicCallback
from aiomisc.service.base import Service
from aiomisc.thread_pool import threaded
from aiomisc.utils import TimeoutType


log = logging.getLogger(__name__)


def _get_watchdog_interval() -> Optional[TimeoutType]:
    value = os.getenv("WATCHDOG_USEC")
    if value is None:
        return None
    return int(value) / 1000000.


def _get_socket_addr() -> Optional[str]:
    addr = os.getenv("NOTIFY_SOCKET")
    if addr is None:
        return None
    if addr[0] == "@":
        addr = "\0" + addr[1:]
    return addr


WATCHDOG_INTERVAL: Optional[TimeoutType] = _get_watchdog_interval()


class SDWatchdogService(Service):
    socket: socket.socket
    socket_addr: Optional[str]
    watchdog_interval: Optional[TimeoutType] = None
    watchdog_timer: PeriodicCallback

    def __init__(
        self,
        watchdog_interval: Optional[TimeoutType] = WATCHDOG_INTERVAL,
        **kwargs: Any,
    ):
        self.watchdog_interval = watchdog_interval
        self.socket_addr = _get_socket_addr()

        super().__init__(**kwargs)

    @property
    def is_connected(self) -> bool:
        return hasattr(self, "socket")

    async def send(self, payload: str) -> None:
        if not self.is_connected:
            return

        try:
            await self.loop.sock_sendall(
                self.socket,
                payload.encode(),
            )
        except (ConnectionError, OSError) as e:
            log.warning("SystemD notify socket communication problem: %r", e)

    @threaded
    def connect(self) -> bool:
        if self.socket_addr is None:
            return False

        self.socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
        self.socket.connect(self.socket_addr)
        self.socket.setblocking(False)
        return True

    @threaded
    def disconnect(self) -> None:
        if not self.is_connected:
            return

        sock = self.socket
        del self.socket

        self.socket_addr = None
        sock.close()

    async def start(self) -> None:
        self.watchdog_timer = PeriodicCallback(self.send, "WATCHDOG=1")

        if self.is_connected and self.watchdog_interval is not None:
            if self.watchdog_interval != WATCHDOG_INTERVAL:
                watchdog_usec = int(self.watchdog_interval * 1000000)
                await self.send(f"WATCHDOG_USEC={watchdog_usec}")

            self.start_event.set()

            # Send notifications twice as often
            self.watchdog_timer.start(self.watchdog_interval / 2)

    async def stop(self, exception: Optional[Exception] = None) -> None:
        await self.watchdog_timer.stop(return_exceptions=True)


def filter_services(
    services: Iterable[Service],
) -> Iterator[SDWatchdogService]:
    for service in services:
        if not isinstance(service, SDWatchdogService):
            continue
        yield service


async def _pre_start(*, services: Tuple[Service, ...], **__: Any) -> None:
    for service in filter_services(services):
        if not service.socket_addr:
            log.debug(
                "NOTIFY_SOCKET not exported. Skipping service %r", service,
            )
            continue

        service.set_loop(get_current_loop())

        if await service.connect():
            await service.send(f"STATUS=Starting {len(services)} services")


async def _post_start(*, services: Tuple[Service, ...], **__: Any) -> None:
    for service in filter_services(services):
        await service.send(f"STATUS=Started {len(services)} services")
        await service.send("READY=1")


async def _pre_stop(*, services: Tuple[Service, ...], **__: Any) -> None:
    for service in filter_services(services):
        await service.send(f"STATUS=Stopping {len(services)} services")


async def _post_stop(*, services: Tuple[Service, ...], **_: Any) -> None:
    for service in filter_services(services):
        await service.send("STOPPING=1")
        await service.disconnect()


def setup() -> None:
    from aiomisc.entrypoint import entrypoint

    entrypoint.PRE_START.connect(_pre_start)
    entrypoint.POST_START.connect(_post_start)
    entrypoint.PRE_STOP.connect(_pre_stop)
    entrypoint.POST_STOP.connect(_post_stop)


__all__ = ("SDWatchdogService",)
__doc__ = "Adds SystemD watchdog support to the entrypoint."



import asyncio
import logging
from types import MappingProxyType
from typing import (
    Any, Awaitable, Callable, Mapping, Optional, Tuple, Type, Union,
)

from aiomisc import Statistic, utils


log = logging.getLogger(__name__)
ExceptionsType = Tuple[Type[Exception], ...]
CallbackType = Callable[..., Union[Awaitable[Any], Any]]
RecurringCallbackStrategy = Callable[
    ["RecurringCallback"],
    Union[int, float, Awaitable[int], Awaitable[float]],
]


class RecurringCallbackStatistic(Statistic):
    call_count: int
    done: int
    fail: int
    sum_time: float


class StrategyException(Exception):
    pass


class StrategyStop(StrategyException):
    """
    Strategy function might raise this exception as way to  stop recurring
    """


class StrategySkip(StrategyException):
    """
    Strategy function might raise this exception as way to skip current call
    """

    def __init__(self, next_attempt_delay: Union[int, float]):
        self.delay = next_attempt_delay


class RecurringCallback:
    __slots__ = ("func", "args", "kwargs", "name", "_statistic")

    def __init__(
        self, coroutine_func: CallbackType,
        *args: Any, **kwargs: Any,
    ):
        self.func: Callable[..., Awaitable[Any]]
        self.args: Tuple[Any, ...]
        self.kwargs: Mapping[str, Any]
        self._statistic: RecurringCallbackStatistic

        self.name: str = repr(coroutine_func)
        self._statistic = RecurringCallbackStatistic(name=self.name)
        self.func = utils.awaitable(coroutine_func)
        self.args = args
        self.kwargs = MappingProxyType(kwargs)

    async def _exec(
        self,
        loop: asyncio.AbstractEventLoop,
        suppress_exceptions: ExceptionsType = (),
    ) -> None:
        self._statistic.call_count += 1
        delta: float = - loop.time()
        try:
            await self.func(*self.args, **self.kwargs)
        except suppress_exceptions:
            self._statistic.fail += 1
            return
        except asyncio.CancelledError:
            raise
        except Exception:
            self._statistic.fail += 1
            log.exception("Recurring task error:")
        else:
            self._statistic.done += 1
        finally:
            delta += loop.time()
            self._statistic.sum_time += delta

    async def _start(
        self,
        strategy: Callable[
            ["RecurringCallback"], Awaitable[Union[int, float]],
        ],
        loop: asyncio.AbstractEventLoop,
        *, shield: bool = False,
        suppress_exceptions: ExceptionsType = (),
    ) -> None:
        runner: Callable[..., Awaitable[Any]]

        while True:
            if loop.is_closed():
                return

            runner = self._exec
            if shield:
                runner = utils.shield(self._exec)

            try:
                delay: Union[int, float] = await strategy(self)
                if not isinstance(delay, (int, float)):
                    log.warning(
                        "Strategy %r returns wrong delay %r. Stopping.",
                        strategy, delay,
                    )
                    return
                if delay < 0:
                    log.warning(
                        "Strategy %r returns negative delay %r. "
                        "Zero delay will be used.",
                        strategy, delay,
                    )
                    delay = 0
            except StrategySkip as e:
                await asyncio.sleep(e.delay)
                continue
            except StrategyException:
                return

            await asyncio.sleep(delay)
            future = loop.create_future()
            task: asyncio.Task = asyncio.ensure_future(
                runner(
                    loop=loop,
                    suppress_exceptions=suppress_exceptions,
                ),
            )

            def on_done(task: asyncio.Task) -> None:
                if future.done():
                    return
                future.set_result(task)

            task.add_done_callback(on_done)

            try:
                await future
            except asyncio.CancelledError:
                if not task.done():
                    task.cancel()
                    await asyncio.gather(task, return_exceptions=True)
                raise

    def start(
        self,
        strategy: RecurringCallbackStrategy,
        loop: Optional[asyncio.AbstractEventLoop] = None, *,
        shield: bool = False,
        suppress_exceptions: ExceptionsType = (),
    ) -> asyncio.Task:
        loop = loop or asyncio.get_event_loop()
        return loop.create_task(
            self._start(
                strategy=utils.awaitable(strategy), loop=loop, shield=shield,
                suppress_exceptions=suppress_exceptions,
            ),
        )

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}({self.name})>"



import asyncio
from contextvars import ContextVar
from typing import Any, Generic, Optional, TypeVar


CT = TypeVar("CT", bound=Any)


class StrictContextVar(Generic[CT]):
    def __init__(self, name: str, exc: Exception):
        self.exc: Exception = exc
        self.context_var: ContextVar = ContextVar(name)

    def get(self) -> CT:
        value: Optional[CT] = self.context_var.get(None)
        if value is None:
            raise self.exc
        return value

    def set(self, value: CT) -> None:
        self.context_var.set(value)


EVENT_LOOP: StrictContextVar[asyncio.AbstractEventLoop] = StrictContextVar(
    "EVENT_LOOP", RuntimeError("no current event loop is set"),
)



import asyncio
import socket
from abc import abstractmethod
from functools import partial
from typing import Any, Awaitable, Callable, Optional, Union

from ..utils import OptionsType, awaitable, bind_socket
from .base import SimpleServer


TransportType = Optional[asyncio.DatagramTransport]
HandleDatagramType = Callable[
    [bytes, tuple], Union[Awaitable[None], None],
]


class UDPServer(SimpleServer):
    class UDPSimpleProtocol(asyncio.DatagramProtocol):

        def __init__(
            self, handle_datagram: HandleDatagramType,
            task_factory: Callable[..., asyncio.Task],
        ):
            super().__init__()
            self.task_factory = task_factory
            self.handler = awaitable(handle_datagram)
            self.transport: TransportType = None
            self.loop: Optional[asyncio.AbstractEventLoop] = None

        def connection_made(self, transport: Any) -> None:
            self.transport = transport
            self.loop = asyncio.get_running_loop()

        def datagram_received(self, data: bytes, addr: tuple) -> None:
            self.task_factory(self.handler(data, addr))

    def __init__(
        self, address: Optional[str] = None, port: Optional[int] = None,
        options: OptionsType = (), sock: Optional[socket.socket] = None,
        **kwargs: Any,
    ):
        if not sock:
            if address is None or port is None:
                raise RuntimeError(
                    "You should pass socket instance or "
                    '"address" and "port" couple',
                )

            self.make_socket = partial(
                bind_socket,
                socket.AF_INET6 if ":" in address else socket.AF_INET,
                socket.SOCK_DGRAM,
                address=address, port=port, options=options,
                proto_name="udp",
            )

        elif not isinstance(sock, socket.socket):
            raise ValueError("sock must be socket instance")
        else:
            self.make_socket = lambda: sock     # type: ignore

        self._transport: TransportType = None
        self._protocol: Optional[asyncio.DatagramProtocol] = None
        self.socket: Optional[socket.socket] = None
        super().__init__(**kwargs)

    def sendto(self, data: bytes, addr: tuple) -> Any:
        if self._transport is None:
            raise RuntimeError

        return self._transport.sendto(data, addr)

    @abstractmethod
    async def handle_datagram(self, data: bytes, addr: tuple) -> None:
        raise NotImplementedError

    async def start(self) -> None:
        self.socket = self.make_socket()

        if self.socket is None or self.loop is None:
            raise RuntimeError

        self._transport, self._protocol = (
            await self.loop.create_datagram_endpoint(   # type: ignore
                lambda: UDPServer.UDPSimpleProtocol(
                    self.handle_datagram,
                    self.create_task,
                ),
                sock=self.socket,
            )
        )

    @property
    def __sockname(self) -> Optional[tuple]:
        if self._transport:
            return self._transport.get_extra_info("sockname")
        return None

    @property
    def address(self) -> Optional[str]:
        if self.__sockname:
            return self.__sockname[0]
        return None

    @property
    def port(self) -> Optional[int]:
        if self.__sockname:
            return self.__sockname[1]
        return None

    async def stop(self, exc: Optional[Exception] = None) -> None:
        await super().stop(exc)
        if self._transport:
            self._transport.close()



import asyncio
import logging
import threading
from collections import Counter, deque
from contextlib import contextmanager
from enum import IntEnum, unique
from functools import wraps
from random import random
from typing import Any, Awaitable, Callable
from typing import Counter as CounterType
from typing import (
    Deque, Generator, Iterable, Optional, Tuple, Type, TypeVar, Union,
)

from aiomisc.compat import EventLoopMixin
from aiomisc.counters import Statistic
from aiomisc.utils import awaitable


log = logging.getLogger(__name__)
Number = Union[int, float]
StatisticType = Deque[Tuple[int, Counter]]
T = TypeVar("T")
ExceptionInspectorType = Optional[Callable[[Exception], bool]]


@unique
class CounterKey(IntEnum):
    FAIL = 0
    OK = 1
    TOTAL = 2


@unique
class CircuitBreakerStates(IntEnum):
    PASSING = 0
    BROKEN = 1
    RECOVERING = 2


class CircuitBreakerStatistic(Statistic):
    call_count: int
    error_ratio: float
    error_ratio_threshold: float
    call_passing: int
    call_broken: int
    call_recovering: int
    call_recovering_ok: int
    call_recovering_failed: int


class CircuitBroken(Exception):
    __slots__ = ("last_exception",)

    def __init__(self, last_exception: Optional[Exception]):
        self.last_exception = last_exception

    def __repr__(self) -> str:
        return f"<{self!r}: {self.last_exception!r}>"


class CircuitBreaker(EventLoopMixin):
    __slots__ = (
        "_broken_time",
        "_counters",
        "_error_ratio",
        "_exceptions",
        "_exception_inspector",
        "_last_exception",
        "_lock",
        "_passing_time",
        "_recovery_at",
        "_recovery_ratio",
        "_recovery_time",
        "_response_time",
        "_state",
        "_statistic",
        "_stuck_until",
    ) + EventLoopMixin.__slots__

    BUCKET_COUNT = 10
    # Thresholds when state will be changed

    # * RECOVER state will be changed to BROKEN
    #   when error ratio will be greater or equal then
    #   RECOVER_BROKEN_THRESHOLD
    RECOVER_BROKEN_THRESHOLD = 0.5
    # * PASSING state will be changed to BROKEN
    #   when error ratio will be greater or equal then
    #   PASSING_BROKEN_THRESHOLD
    PASSING_BROKEN_THRESHOLD = 1

    _stuck_until: Number
    _recovery_at: Number
    _last_exception: Optional[Exception]

    def __init__(
        self,
        error_ratio: float,
        response_time: Number,
        exceptions: Iterable[Type[Exception]] = (Exception,),
        recovery_time: Optional[Number] = None,
        broken_time: Optional[Number] = None,
        passing_time: Optional[Number] = None,
        exception_inspector: Optional[ExceptionInspectorType] = None,
        statistic_name: Optional[str] = None,
    ):
        """
        Circuit Breaker pattern implementation. The class instance collects
        call statistics through the ``call`` or ``call async`` methods.

        The state machine has three states:
        * ``CircuitBreakerStates.PASSING``
        * ``CircuitBreakerStates.BROKEN``
        * ``CircuitBreakerStates.RECOVERING``

        In passing mode all results or exceptions will be returned as is.
        Statistic collects for each call.

        In broken mode returns exception ``CircuitBroken`` for each call.
        Statistic doesn't collecting.

        In recovering mode the part of calls is real function calls and
        remainings raises ``CircuitBroken``. The count of real calls grows
        exponentially in this case but when 20% (by default) will be failed
        the state returns to broken state.

        :param error_ratio: Failed to success calls ratio. The state might be
                            changed if ratio will reach given value within
                            ``response time`` (in seconds).
                            Value between 0.0 and 1.0.
        :param response_time: Time window to collect statistics (seconds)
        :param exceptions: Only this exceptions will affect ratio.
                           Base class  ``Exception`` used by default.
        :param recovery_time: minimal time in recovery state (seconds)
        :param broken_time: minimal time in broken state (seconds)
        :param passing_time: minimum time in passing state (seconds)
        """
        if response_time <= 0:
            raise ValueError("Response time must be greater then zero")

        if 0. > error_ratio >= 1.:
            raise ValueError(
                "Error ratio must be between 0 and 1 not %r" % error_ratio,
            )

        self._statistic: StatisticType = deque(
            maxlen=self.BUCKET_COUNT,
        )
        self._lock = threading.RLock()
        self._error_ratio = error_ratio
        self._state = CircuitBreakerStates.PASSING
        self._response_time = response_time
        self._stuck_until = 0
        self._recovery_at = 0

        self._exceptions = tuple(frozenset(exceptions))
        self._exception_inspector = exception_inspector

        self._passing_time = passing_time or self._response_time
        self._broken_time = broken_time or self._response_time
        self._recovery_time = recovery_time or self._response_time
        self._last_exception = None
        self._counters = CircuitBreakerStatistic(statistic_name)
        self._counters.error_ratio_threshold = error_ratio

    @property
    def response_time(self) -> Number:
        return self._response_time

    @property
    def state(self) -> CircuitBreakerStates:
        return self._state

    def _get_time(self) -> float:
        return self.loop.time()

    def bucket(self) -> int:
        ts = self._get_time() * self.BUCKET_COUNT
        return int(ts - (ts % self._response_time))

    def counter(self) -> Counter:
        with self._lock:
            current = self.bucket()

            if not self._statistic:
                # Empty statistic just return a new counter
                counter: CounterType[int] = Counter()
                self._statistic.append((current, counter))
                return counter

            bucket, counter = self._statistic[-1]

            if current != bucket:
                # Append Counter to statistic or shift when maxsize reached
                counter = Counter()
                self._statistic.append((current, counter))

            return counter

    def __gen_statistic(self) -> Generator[Counter, None, None]:
        """
        Generator which returns only buckets Counters not before current_time
        """
        not_before = self.bucket() - (self._response_time * self.BUCKET_COUNT)

        for idx in range(len(self._statistic) - 1, -1, -1):
            bucket, counter = self._statistic[idx]

            if bucket < not_before:
                break

            yield counter

    def get_state_delay(self) -> Number:
        delay = self._stuck_until - self._get_time()
        return max(delay, 0)

    def _inspect_exception(self, e: Exception) -> int:
        if not self._exception_inspector:
            return 1

        # noinspection PyBroadException
        try:
            return 1 if self._exception_inspector(e) else 0
        except Exception:
            log.exception(
                "Unhandled exception in %r",
                self._exception_inspector,
            )
            return 1

    def _on_passing(
        self, counter: CounterType[int],
    ) -> Generator[Any, Any, Any]:
        try:
            yield
            counter[CounterKey.OK] += 1
            self._last_exception = None
        except self._exceptions as e:
            self._last_exception = e
            counter[CounterKey.FAIL] += self._inspect_exception(e)
            raise
        finally:
            counter[CounterKey.TOTAL] += 1

    def _on_recover(
        self, counter: CounterType[int],
    ) -> Generator[Any, Any, Any]:
        current_time = self._get_time()
        condition = (random() + 1) < (
            2 ** ((current_time - self._recovery_at) / self._recovery_time)
        )

        if not condition:
            self._counters.call_recovering_failed += 1
            raise CircuitBroken(self._last_exception)

        self._counters.call_recovering_ok += 1
        yield from self._on_passing(counter)

    @property
    def recovery_ratio(self) -> Number:
        total_count = 0
        upper_count = 0

        for counter in self.__gen_statistic():
            total_count += 1

            if not counter[CounterKey.TOTAL]:
                continue

            fail_ratio = counter[CounterKey.FAIL] / counter[CounterKey.TOTAL]

            if fail_ratio >= self._error_ratio:
                upper_count += 1

        if not total_count:
            return 0

        return upper_count / total_count

    def _compute_state(self) -> None:
        current_time = self._get_time()

        if current_time < self._stuck_until:
            # Skip state changing until
            return

        if self._state is CircuitBreakerStates.BROKEN:
            self._state = CircuitBreakerStates.RECOVERING
            self._recovery_at = current_time
            return

        # Do not compute when not enough statistic
        if (
            self._state is CircuitBreakerStates.PASSING
            and len(self._statistic) < self.BUCKET_COUNT
        ):
            return

        recovery_ratio = self.recovery_ratio

        if self._state is CircuitBreakerStates.PASSING:
            if recovery_ratio >= self.PASSING_BROKEN_THRESHOLD:
                self._stuck_until = current_time + self._broken_time
                self._state = CircuitBreakerStates.BROKEN
                self._statistic.clear()
            return

        if self._state is not CircuitBreakerStates.RECOVERING:
            return

        if recovery_ratio >= (
            self.RECOVER_BROKEN_THRESHOLD * self._error_ratio
        ):
            self._stuck_until = current_time + self._broken_time
            self._state = CircuitBreakerStates.BROKEN
            self._statistic.clear()
            return

        recovery_length = current_time - self._recovery_at
        if recovery_length >= self._recovery_time:
            self._stuck_until = current_time + self._passing_time
            self._state = CircuitBreakerStates.PASSING
            return

    @contextmanager
    def context(self) -> Generator[Any, Any, Any]:
        counter = self.counter()
        self._compute_state()
        self._counters.call_count += 1

        if self._state is CircuitBreakerStates.PASSING:
            self._counters.call_passing += 1
            yield from self._on_passing(counter)
            return

        elif self._state is CircuitBreakerStates.BROKEN:
            self._counters.call_broken += 1
            raise CircuitBroken(self._last_exception)

        elif self._state is CircuitBreakerStates.RECOVERING:
            self._counters.call_recovering += 1
            yield from self._on_recover(counter)
            return

        raise NotImplementedError(self._state)

    def call(
        self, func: Callable[..., T], *args: Any, **kwargs: Any,
    ) -> T:
        with self.context():
            return func(*args, **kwargs)

    async def call_async(
        self, func: Callable[..., Awaitable[T]],
        *args: Any, **kwargs: Any,
    ) -> T:
        with self.context():
            return await awaitable(func)(*args, **kwargs)   # type: ignore

    def __repr__(self) -> str:
        return "<{}: state={!r} recovery_ratio={!s}>".format(
            self.__class__.__name__, self._state, self.recovery_ratio,
        )


CutoutFuncType = Union[Callable[..., T], Callable[..., Awaitable[T]]]
CutoutDecoratorReturnType = Callable[..., Union[T, Awaitable[T]]]
CutoutReturnType = Callable[[CutoutFuncType], CutoutDecoratorReturnType]


def cutout(
    ratio: float, response_time: Union[int, float],
    *exceptions: Type[Exception], **kwargs: Any,
) -> CutoutReturnType:
    circuit_breaker = CircuitBreaker(
        error_ratio=ratio,
        response_time=response_time,
        exceptions=exceptions,
        **kwargs,
    )

    def decorator(func: CutoutFuncType) -> CutoutDecoratorReturnType:
        @wraps(func)
        async def async_wrapper(
            *args: Any, **kw: Any,
        ) -> T:
            return await circuit_breaker.call_async(func, *args, **kw)

        @wraps(func)
        def wrapper(*args: Any, **kw: Any) -> Any:
            return circuit_breaker.call(func, *args, **kw)

        if asyncio.iscoroutinefunction(func):
            return async_wrapper

        return wrapper

    return decorator



import logging
import sys
from typing import IO, Any, Optional

from aiomisc_log.enum import DateFormat


try:
    from rich.console import Console
    from rich.logging import RichHandler

    def rich_formatter(
        date_format: Optional[str] = None, stream: Optional[IO[str]] = None,
        rich_tracebacks: bool = False, **_: Any,
    ) -> logging.Handler:
        handler = RichHandler(
            console=Console(file=stream or sys.stderr),
            log_time_format=date_format or DateFormat.rich.value,
            rich_tracebacks=rich_tracebacks,
        )
        formatter = logging.Formatter("%(message)s")
        handler.setFormatter(formatter)
        return handler
except ImportError:
    def rich_formatter(
        date_format: Optional[str] = None, stream: Optional[IO[str]] = None,
        rich_tracebacks: bool = False, **_: Any,
    ) -> logging.Handler:
        raise ImportError("You must install \"rich\" library for use it")



import asyncio
import bz2
import gzip
import lzma
import sys
from concurrent.futures import Executor
from enum import Enum
from functools import partial, total_ordering
from pathlib import Path
from typing import (
    IO, Any, AnyStr, Awaitable, Callable, Generator, Generic, List, Optional,
    TextIO, TypeVar, Union,
)

from .compat import EventLoopMixin


T = TypeVar("T", bound=Any)


def proxy_method_async(
    name: str,
    in_executor: bool = True,
) -> Callable[..., Any]:
    def wrap_to_future(
        loop: asyncio.AbstractEventLoop,
        func: Callable[..., T],
        *args: Any, **kwargs: Any,
    ) -> asyncio.Future:
        future = loop.create_future()

        def _inner():   # type: ignore
            try:
                return future.set_result(func(*args, **kwargs))
            except Exception as e:
                return future.set_exception(e)

        loop.call_soon(_inner)
        return future

    def wrap_to_thread(
        loop: asyncio.AbstractEventLoop, func: Callable[..., T],
        executor: Executor,
        *args: Any, **kwargs: Any,
    ) -> Awaitable[T]:
        callee = partial(func, *args, **kwargs)
        # noinspection PyTypeChecker
        return loop.run_in_executor(executor, callee)

    async def method(self, *args, **kwargs):    # type: ignore
        func = getattr(self.fp, name)

        if in_executor:
            return await wrap_to_thread(
                self.loop, func, self.executor, *args, **kwargs,
            )

        return await wrap_to_future(self.loop, func, *args, **kwargs)

    method.__name__ = name
    return method


@total_ordering
class AsyncFileIO(EventLoopMixin, Generic[AnyStr]):
    __slots__ = (
        "_fp", "executor", "__iterator_lock",
    ) + EventLoopMixin.__slots__

    _fp: Optional[IO[AnyStr]]

    @staticmethod
    def get_opener() -> Callable[..., IO[AnyStr]]:
        return open

    def __init__(
        self, fname: Union[str, Path], mode: str = "r",
        executor: Optional[Executor] = None, *args: Any,
        loop: Optional[asyncio.AbstractEventLoop] = None, **kwargs: Any,
    ):
        self.executor = executor
        self._loop = loop
        self._fp = None
        self.__open_args = (str(fname), mode, *args), kwargs
        self.__iterator_lock = asyncio.Lock()

    @classmethod
    def open_fp(
        cls, fp: IO, executor: Optional[Executor] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ) -> "AsyncFileIO":
        async_fp = cls(fp.name, mode=fp.mode, executor=executor, loop=loop)
        async_fp._fp = fp
        return async_fp

    async def open(self) -> None:
        if self._fp is not None:
            return

        args, kwargs = self.__open_args
        self._fp = await self.__execute_in_thread(
            self.get_opener(), *args, **kwargs,
        )

    @property
    def fp(self) -> IO[AnyStr]:
        if self._fp is not None:
            return self._fp
        raise RuntimeError("file is not opened")

    def closed(self) -> bool:
        return self.fp.closed

    def __await__(self) -> Generator[Any, Any, "AsyncFileIO"]:
        yield from self.open().__await__()
        return self

    async def __aenter__(self) -> "AsyncFileIO[AnyStr]":
        await self.open()
        return self

    async def __aexit__(
        self, exc_type: Any,
        exc_val: Any, exc_tb: Any,
    ) -> None:
        if self.fp is None:
            raise RuntimeError("file is not opened")

        await self.loop.run_in_executor(
            None, self.fp.__exit__, exc_type, exc_val, exc_tb,
        )

    def __del__(self) -> None:
        if not self.fp or self.fp.closed:
            return

        self.fp.close()
        del self._fp

    def __aiter__(self) -> "AsyncFileIO[AnyStr]":
        return self

    async def __anext__(self) -> AnyStr:
        async with self.__iterator_lock:
            line = await self.readline()

        if not len(line):
            raise StopAsyncIteration

        return line

    def __eq__(self, other: Any) -> bool:
        return (
            self.__class__, self.fp.__eq__(other),
        ) == (
            other.__class__, self.fp.__eq__(other),
        )

    def __lt__(self, other: Any) -> bool:
        return self.fp < other.fp

    def __hash__(self) -> int:
        return hash((self.__class__, self.fp))

    async def __execute_in_thread(
        self, method: Callable[..., Any], *args: Any, **kwargs: Any,
    ) -> Any:
        return await self.loop.run_in_executor(
            self.executor, partial(method, *args, **kwargs),
        )

    @property
    def mode(self) -> str:
        return self.fp.mode

    @property
    def name(self) -> str:
        return self.fp.name

    def fileno(self) -> int:
        return self.fp.fileno()

    def isatty(self) -> bool:
        return self.fp.isatty()

    async def close(self) -> None:
        return await self.__execute_in_thread(self.fp.close)

    def __getattr__(self, name: str) -> Callable[..., Awaitable[Any]]:
        async def method(*args: Any) -> Any:
            getter = getattr(self.fp, name)
            if callable(getter):
                return await self.__execute_in_thread(getter, *args)
            return getter
        method.__name__ = name
        return method

    async def tell(self) -> int:
        return self.fp.tell()

    async def readable(self) -> bool:
        return self.fp.readable()

    async def seekable(self) -> bool:
        return self.fp.seekable()

    async def writable(self) -> bool:
        return self.fp.writable()

    async def flush(self) -> None:
        await self.__execute_in_thread(self.fp.flush)

    async def read(self, n: int = -1) -> AnyStr:
        return await self.__execute_in_thread(self.fp.read, n)

    async def readline(self, limit: int = -1) -> AnyStr:
        return await self.__execute_in_thread(self.fp.readline, limit)

    async def readlines(self, limit: int = -1) -> List[AnyStr]:
        return await self.__execute_in_thread(self.fp.readlines, limit)

    async def seek(self, offset: int, whence: int = 0) -> int:
        return await self.__execute_in_thread(self.fp.seek, offset, whence)

    async def truncate(self, size: Optional[int] = None) -> int:
        return await self.__execute_in_thread(self.fp.truncate, size)

    async def write(self, s: AnyStr) -> int:
        return await self.__execute_in_thread(self.fp.write, s)

    async def writelines(self, lines: List[AnyStr]) -> None:
        await self.__execute_in_thread(self.fp.writelines, lines)


AsyncFileIOBase = AsyncFileIO


class AsyncBinaryIO(AsyncFileIO[bytes]):
    pass


class AsyncTextIO(AsyncFileIO[str]):
    @property
    def fp(self) -> TextIO:
        return self._fp     # type: ignore

    @property
    def newlines(self) -> Any:
        return self.fp.newlines

    @property
    def errors(self) -> Optional[str]:
        return self.fp.errors

    @property
    def line_buffering(self) -> int:
        return self.fp.line_buffering

    @property
    def encoding(self) -> str:
        return self.fp.encoding

    def buffer(self) -> AsyncBinaryIO:
        return AsyncBinaryIO.open_fp(self.fp.buffer)    # type: ignore


# Aliases
AsyncBytesFileIO = AsyncBinaryIO
AsyncTextFileIO = AsyncTextIO


class AsyncGzipBinaryIO(AsyncBytesFileIO):
    @staticmethod
    def get_opener() -> Callable[..., IO[AnyStr]]:
        return gzip.open  # type: ignore


class AsyncGzipTextIO(AsyncTextFileIO):
    @staticmethod
    def get_opener() -> Callable[..., IO[AnyStr]]:
        return gzip.open  # type: ignore


class AsyncBz2BinaryIO(AsyncBytesFileIO):
    @staticmethod
    def get_opener() -> Callable[..., IO[AnyStr]]:
        return bz2.open   # type: ignore


class AsyncBz2TextIO(AsyncTextFileIO):
    @staticmethod
    def get_opener() -> Callable[..., IO[AnyStr]]:
        return bz2.open   # type: ignore


class AsyncLzmaBinaryIO(AsyncBytesFileIO):
    @staticmethod
    def get_opener() -> Callable[..., IO[AnyStr]]:
        return lzma.open  # type: ignore


class AsyncLzmaTextIO(AsyncTextFileIO):
    @staticmethod
    def get_opener() -> Callable[..., IO[AnyStr]]:
        return lzma.open  # type: ignore


class Compression(Enum):
    NONE = (AsyncBinaryIO, AsyncTextIO)
    GZIP = (AsyncGzipBinaryIO, AsyncGzipTextIO)
    BZ2 = (AsyncBz2BinaryIO, AsyncBz2TextIO)
    LZMA = (AsyncLzmaBinaryIO, AsyncLzmaTextIO)


AsyncFileType = Union[AsyncFileIO[AnyStr], AsyncTextIO, AsyncBinaryIO]

# Deprecated excluded from __all__
AsyncFileT = AsyncFileType


def async_open(
    fname: Union[str, Path], mode: str = "r",
    compression: Compression = Compression.NONE,
    encoding: str = sys.getdefaultencoding(),
    *args: Any, **kwargs: Any,
) -> AsyncFileType:
    binary_io_class, text_io_class = compression.value

    if "b" in mode:
        return binary_io_class(fname, mode, *args, **kwargs)

    if "t" not in mode:
        mode = f"t{mode}"

    return text_io_class(fname, mode, encoding=encoding, *args, **kwargs)


__all__ = (
    "AsyncBinaryIO",
    "AsyncBz2BinaryIO",
    "AsyncBz2TextIO",
    "AsyncFileIO",
    "AsyncFileType",
    "AsyncGzipBinaryIO",
    "AsyncGzipTextIO",
    "AsyncLzmaBinaryIO",
    "AsyncLzmaTextIO",
    "AsyncTextIO",
    "Compression",
    "async_open",
)



import asyncio
import itertools
import logging.handlers
import socket
import uuid
from functools import wraps
from random import getrandbits
from typing import (
    Any, Awaitable, Callable, Collection, Coroutine, Generator, Iterable,
    Iterator, List, Optional, Set, Tuple, TypeVar, Union,
)

from .compat import (
    event_loop_policy, sock_set_nodelay, sock_set_reuseport, time_ns,
)
from .thread_pool import ThreadPoolExecutor


T = TypeVar("T", bound=Any)
TimeoutType = Union[int, float]


log = logging.getLogger(__name__)


def fast_uuid4() -> uuid.UUID:
    """ Fast UUID4 like identifier """
    return uuid.UUID(int=getrandbits(128), version=4)


__NODE = uuid.getnode()


def fast_uuid1() -> uuid.UUID:
    """ Fast UUID1 like identifier """
    value = time_ns()
    value = (value << 16) + getrandbits(16)
    value = (value << 48) + __NODE
    return uuid.UUID(int=value, version=1)


def chunk_list(iterable: Iterable[T], size: int) -> Iterable[List[T]]:
    """
    Split list or generator by chunks with fixed maximum size.
    """

    iterable = iter(iterable)

    item = list(itertools.islice(iterable, size))
    while item:
        yield item
        item = list(itertools.islice(iterable, size))


OptionsType = Iterable[Tuple[int, int, int]]


def bind_socket(
    *args: Any, address: str, port: int = 0, options: OptionsType = (),
    reuse_addr: bool = True, reuse_port: bool = True,
    proto_name: Optional[str] = None,
) -> socket.socket:
    """

    Bind socket and set ``setblocking(False)`` for just created socket.
    This detects ``address`` format and select socket family automatically.

    :param args: which will be passed to stdlib's socket constructor (optional)
    :param address: bind address
    :param port: bind port
    :param options: Tuple of pairs which contain socket option
                    to set and the option value.
    :param reuse_addr: set socket.SO_REUSEADDR
    :param reuse_port: set socket.SO_REUSEPORT
    :param proto_name: protocol name which will be logged after binding
    :return: socket.socket
    """

    if not args:
        if ":" in address:
            args = (socket.AF_INET6, socket.SOCK_STREAM)
        else:
            args = (socket.AF_INET, socket.SOCK_STREAM)

    sock = socket.socket(*args)
    sock.setblocking(False)

    if not args and ":" in address:
        sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)

    if reuse_addr:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    if reuse_port:
        sock_set_reuseport(sock, True)

    sock_set_nodelay(sock)

    for level, option, value in options:
        sock.setsockopt(level, option, value)

    unix_address_family = getattr(socket, "AF_UNIX", None)
    if sock.family == unix_address_family:
        proto_name = proto_name or "unix"
        sock.bind(address)
    else:
        proto_name = proto_name or "tcp"
        sock.bind((address, port))

    sock_addr = sock.getsockname()
    if not isinstance(sock_addr, str):
        sock_addr = sock_addr[:2]

    if sock.family == socket.AF_INET6:
        log.info("Listening %s://[%s]:%s", proto_name, *sock_addr)
    elif sock.family == unix_address_family:
        log.info("Listening %s://%s", proto_name, sock_addr)
    else:
        log.info("Listening %s://%s:%s", proto_name, *sock_addr)

    return sock


def create_default_event_loop(
    pool_size: Optional[int] = None,
    policy: asyncio.AbstractEventLoopPolicy = event_loop_policy,
    debug: bool = False,
) -> Tuple[asyncio.AbstractEventLoop, ThreadPoolExecutor]:
    """
    Creates an event loop and thread pool executor

    :param pool_size: thread pool maximal size
    :param policy: event loop policy
    :param debug: set ``loop.set_debug(True)`` if True
    """

    current_loop_is_running = False
    try:
        current_loop = asyncio.get_event_loop()
        current_loop_is_running = current_loop.is_running()
        del current_loop
    except RuntimeError:
        pass

    if current_loop_is_running:
        raise RuntimeError(
            "Trying to create new event loop instance but another "
            "default loop in this thread is running right now.",
        )

    asyncio.set_event_loop_policy(policy)

    loop = asyncio.new_event_loop()
    loop.set_debug(debug)
    asyncio.set_event_loop(loop)

    pool_size = pool_size or ThreadPoolExecutor.DEFAULT_POOL_SIZE
    thread_pool = ThreadPoolExecutor(pool_size, statistic_name="default")
    loop.set_default_executor(thread_pool)

    return loop, thread_pool


def new_event_loop(
    pool_size: Optional[int] = None,
    policy: asyncio.AbstractEventLoopPolicy = event_loop_policy,
) -> asyncio.AbstractEventLoop:
    loop, thread_pool = create_default_event_loop(pool_size, policy)
    return loop


def shield(
    func: Callable[..., Coroutine[Any, Any, T]],
) -> Callable[..., Coroutine[Any, Any, T]]:
    """
    Simple and useful decorator for wrap the coroutine to `asyncio.shield`.

    >>> @shield
    ... async def non_cancelable_func():
    ...     await asyncio.sleep(1)

    """

    async def awaiter(future: Awaitable[Any]) -> Any:
        return await future

    @wraps(func)
    def wrap(*args: Any, **kwargs: Any) -> Coroutine[Any, Any, T]:
        return wraps(func)(awaiter)(asyncio.shield(func(*args, **kwargs)))

    return wrap


class SelectResult(Collection):
    __slots__ = ("length", "result_idx", "is_exception", "value")

    def __init__(self, length: int):
        self.length = length
        self.result_idx = None      # type: Optional[int]
        self.is_exception = None    # type: Optional[bool]
        self.value = None           # type: Any

    def __len__(self) -> int:
        return self.length

    def __contains__(self, x: Any) -> bool:
        return self.value is x

    def set_result(self, idx: int, value: Any, is_exception: bool) -> None:
        if self.result_idx is not None:
            return

        self.value = value
        self.result_idx = idx
        self.is_exception = is_exception

    def result(self) -> Any:
        if self.is_exception:
            raise self.value
        return self.value

    def done(self) -> bool:
        return self.result_idx is not None

    def __iter__(self) -> Iterator[Optional[T]]:
        for i in range(self.length):
            yield self.value if i == self.result_idx else None


class SelectAwaitable:
    """
    Select one of passed awaitables
    """

    __slots__ = (
        "_awaitables", "_return_exceptions", "_timeout", "_wait",
        "__loop", "_cancel", "_result",
    )

    _result: SelectResult

    def __init__(
        self, *awaitables: Awaitable[T],
        return_exceptions: bool = False, cancel: bool = True,
        timeout: Optional[TimeoutType] = None,
        wait: bool = True, loop: Optional[asyncio.AbstractEventLoop] = None,
    ):
        """
        :param awaitables: awaitable objects
        :param return_exceptions: if True exception will not be raised
                                  just returned as result
        :param cancel: cancel unfinished coroutines (default True)
        :param timeout: execution timeout
        :param wait: when False and ``cancel=True``, unfinished coroutines will
                     be cancelled in the background.
        :param loop: event loop
        """
        self.__loop = loop
        self._awaitables = awaitables
        self._cancel = cancel
        self._return_exceptions = return_exceptions
        self._timeout = timeout
        self._wait = wait

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        if self.__loop is None:
            self.__loop = asyncio.get_running_loop()
        return self.__loop

    async def __waiter(self, idx: int, awaitable: Awaitable[T]) -> None:
        try:
            ret = await awaitable
        except asyncio.CancelledError:
            raise
        except Exception as e:
            return self._result.set_result(idx, e, is_exception=True)
        self._result.set_result(idx, ret, is_exception=False)

    async def __run(
        self, coroutines: Iterable[asyncio.Future],
    ) -> SelectResult:
        try:
            _, pending = await asyncio.wait(
                coroutines,
                timeout=self._timeout,
                return_when=asyncio.FIRST_COMPLETED,
            )

            if self._cancel:
                cancelling = cancel_tasks(pending)

                if self._wait:
                    await cancelling

            if self._result.is_exception and not self._return_exceptions:
                self._result.result()
            return self._result
        except TimeoutError as e:
            self._result.set_result(0, e, is_exception=True)
            raise
        except asyncio.CancelledError:
            await cancel_tasks(coroutines)
            raise

    def __await__(self) -> Generator[Any, None, SelectResult]:
        self._result = SelectResult(len(self._awaitables))
        coroutines = [
            asyncio.ensure_future(self.__waiter(idx, coroutine))
            for idx, coroutine in enumerate(self._awaitables)
        ]
        # Prevent double __await__ call
        del self._awaitables
        return self.__run(coroutines).__await__()


select = SelectAwaitable


def pending_futures(
    futures: Iterable[asyncio.Future],
) -> Iterator[asyncio.Future]:
    # Copying collection to ignore it
    # changes during iteration
    for future in tuple(futures):
        if future.done():
            continue
        yield future


def set_exception(
    futures: Iterable[asyncio.Future],
    exc: Union[BaseException] = asyncio.CancelledError(),
) -> Set[asyncio.Task]:
    cancelled_tasks = set()

    for future in pending_futures(futures):
        if isinstance(future, asyncio.Task):
            future.cancel()
            cancelled_tasks.add(future)
        elif isinstance(future, asyncio.Future):
            future.set_exception(exc)
        else:
            log.warning(
                "Skipping object %r because it's not a Task or Future",
                future,
            )
    return cancelled_tasks


def cancel_tasks(tasks: Iterable[asyncio.Future]) -> asyncio.Future:
    """
    All passed tasks will be cancelled and a new task will be returned.

    :param tasks: tasks which will be cancelled
    """

    future = asyncio.get_event_loop().create_future()
    future.set_result(None)

    if not tasks:
        return future

    exc = asyncio.CancelledError()
    cancelled_tasks = set_exception(tasks, exc)

    if not cancelled_tasks:
        return future

    waiter = asyncio.ensure_future(
        asyncio.gather(
            *cancelled_tasks, return_exceptions=True,
        ),
    )

    return waiter


AT = TypeVar("AT", bound=Any)


def awaitable(
    func: Callable[..., Union[AT, Awaitable[AT]]],
) -> Callable[..., Awaitable[AT]]:
    """

    Decorator wraps function and returns a function which returns
    awaitable object. In case than a function returns a future,
    the original future will be returned. In case then the function
    returns a coroutine, the original coroutine will be returned.
    In case than function returns non-awaitable object, it's will
    be wrapped to a new coroutine which just returns this object.
    It's useful when you don't want to check function result before
    use it in ``await`` expression.
    """

    # Avoid python 3.8+ warning
    if asyncio.iscoroutinefunction(func):
        return func

    async def awaiter(obj: AT) -> AT:
        return obj

    @wraps(func)
    def wrap(*args: Any, **kwargs: Any) -> Awaitable[AT]:
        result = func(*args, **kwargs)

        if hasattr(result, "__await__"):
            return result
        if asyncio.iscoroutine(result) or asyncio.isfuture(result):
            return result

        return awaiter(result)      # type: ignore

    return wrap



from collections import Counter
from typing import (
    Any, Dict, FrozenSet, Generator, MutableMapping, MutableSet, NamedTuple,
    Optional, Set, Tuple, Type, Union,
)
from weakref import WeakSet


class Metric:
    __slots__ = ("name", "counter")

    def __init__(
        self, name: str,
        counter: MutableMapping[str, Union[float, int]],
        default: Union[float, int] = 0,
    ):
        self.name: str = name
        self.counter = counter
        self.counter[name] = default

    def __get__(self) -> Union[float, int]:
        return self.counter[self.name]

    def __set__(self, value: Union[float, int]) -> None:
        self.counter[self.name] = value

    def __iadd__(self, value: Union[float, int]) -> "Metric":
        self.counter[self.name] += value
        return self

    def __isub__(self, value: Union[float, int]) -> "Metric":
        self.counter[self.name] -= value
        return self

    def __eq__(self, other: Any) -> bool:
        return self.counter[self.name] == other

    def __hash__(self) -> int:
        return hash(self.counter[self.name])


class AbstractStatistic:
    __metrics__: FrozenSet[str]
    __instances__: MutableSet["AbstractStatistic"]
    _counter: MutableMapping[str, Union[float, int]]
    name: Optional[str]


CLASS_STORE: Set[Type[AbstractStatistic]] = set()


class MetaStatistic(type):
    def __new__(
        mcs, name: str,
        bases: Tuple[type, ...],
        dct: Dict[str, Any],
    ) -> Any:

        # noinspection PyTypeChecker
        klass: Type[AbstractStatistic] = super().__new__(  # type: ignore
            mcs, name, bases, dct,
        )

        metrics = set()

        for base_class in bases:
            if not issubclass(base_class, AbstractStatistic):
                continue

            if not hasattr(base_class, "__annotations__"):
                continue

            for prop, kind in base_class.__annotations__.items():
                if kind not in (int, float):
                    continue

                if prop.startswith("_"):
                    continue

                metrics.add(prop)

        for prop, kind in klass.__annotations__.items():
            if kind not in (int, float):
                continue

            metrics.add(prop)

        klass.__metrics__ = frozenset(metrics)

        if klass.__metrics__:
            klass.__instances__ = WeakSet()
            CLASS_STORE.add(klass)

        return klass


class Statistic(AbstractStatistic, metaclass=MetaStatistic):
    __slots__ = ("_counter", "name")

    def __init__(self, name: Optional[str] = None) -> None:
        self._counter = Counter()   # type: ignore
        self.name = name

        for prop in self.__metrics__:
            setattr(self, prop, Metric(prop, self._counter))

        self.__instances__.add(self)


class StatisticResult(NamedTuple):
    kind: Type[AbstractStatistic]
    name: Optional[str]
    metric: str
    value: Union[int, float]


# noinspection PyProtectedMember
def get_statistics(
    *kind: Type[Statistic],
) -> Generator[Any, Tuple[Statistic, str, int], None]:
    for klass in CLASS_STORE:
        if kind and not issubclass(klass, kind):
            continue

        for instance in klass.__instances__:
            for metric, value in instance._counter.items():
                yield StatisticResult(
                    kind=klass,
                    name=instance.name,
                    metric=metric,
                    value=value,
                )



import asyncio
from functools import wraps
from typing import (
    Any, Awaitable, Callable, Optional, Tuple, Type, TypeVar, Union,
)

from .counters import Statistic
from .timeout import timeout


Number = Union[int, float]
T = TypeVar("T")


WrapReturnType = Callable[..., Awaitable[T]]
ReturnType = Callable[..., WrapReturnType]


class BackoffStatistic(Statistic):
    done: int
    attempts: int
    cancels: int
    errors: int
    sum_time: float


class RetryStatistic(BackoffStatistic):
    pass


# noinspection SpellCheckingInspection
def asyncbackoff(
    attempt_timeout: Optional[Number],
    deadline: Optional[Number],
    pause: Number = 0,
    *exc: Type[Exception], exceptions: Tuple[Type[Exception], ...] = (),
    max_tries: Optional[int] = None,
    giveup: Optional[Callable[[Exception], bool]] = None,
    statistic_name: Optional[str] = None,
    statistic_class: Type[BackoffStatistic] = BackoffStatistic,
) -> ReturnType:
    """
    Patametric decorator that ensures that ``attempt_timeout`` and
    ``deadline`` time limits are met by decorated function.

    In case of exception function will be called again with similar
    arguments after ``pause`` seconds.

    :param statistic_name: name filed for statistic instances
    :param attempt_timeout: is maximum execution time for one
                            execution attempt.
    :param deadline: is maximum execution time for all execution attempts.
    :param pause: is time gap between execution attempts.
    :param exc: retrying when this exceptions was raised.
    :param exceptions: similar as exc but keyword only.
    :param max_tries: is maximum count of execution attempts (>= 1).
    :param giveup: is a predicate function which can decide by a given
    :param statistic_class: statistic class
    """

    exceptions = exc + tuple(exceptions)
    statistic = statistic_class(statistic_name)

    if not pause:
        pause = 0
    elif pause < 0:
        raise ValueError("'pause' must be positive")

    if attempt_timeout is not None and attempt_timeout < 0:
        raise ValueError("'attempt_timeout' must be positive or None")

    if deadline is not None and deadline < 0:
        raise ValueError("'deadline' must be positive or None")

    if max_tries is not None and max_tries < 1:
        raise ValueError("'max_retries' must be >= 1 or None")

    if giveup is not None and not callable(giveup):
        raise ValueError("'giveup' must be a callable or None")

    exceptions = tuple(exceptions) or ()
    exceptions += asyncio.TimeoutError,

    def decorator(func: WrapReturnType) -> WrapReturnType:
        if attempt_timeout is not None:
            func = timeout(attempt_timeout)(func)

        @wraps(func)
        async def wrap(*args: Any, **kwargs: Any) -> T:
            last_exc = None
            tries = 0

            async def run() -> Any:
                nonlocal last_exc, tries

                loop = asyncio.get_running_loop()

                while True:
                    statistic.attempts += 1
                    tries += 1
                    delta = -loop.time()

                    try:
                        return await asyncio.wait_for(
                            func(*args, **kwargs),
                            timeout=attempt_timeout,
                        )
                    except asyncio.CancelledError:
                        statistic.cancels += 1
                        raise
                    except exceptions as e:
                        statistic.errors += 1
                        last_exc = e
                        if max_tries is not None and tries >= max_tries:
                            raise
                        if giveup and giveup(e):
                            raise
                        await asyncio.sleep(pause)
                    except Exception as e:
                        last_exc = e
                        raise
                    finally:
                        delta += loop.time()
                        statistic.sum_time += delta
                        statistic.done += 1

            try:
                return await asyncio.wait_for(run(), timeout=deadline)
            except Exception:
                if last_exc:
                    raise last_exc
                raise

        return wrap
    return decorator


def asyncretry(
    max_tries: Optional[int],
    exceptions: Tuple[Type[Exception], ...] = (Exception,),
    pause: Number = 0,
    giveup: Optional[Callable[[Exception], bool]] = None,
    statistic_name: Optional[str] = None,
) -> ReturnType:
    """
    Shortcut of ``asyncbackoff(None, None, 0, Exception)``.

    In case of exception function will be called again with similar
    arguments after ``pause`` seconds.

    :param max_tries: is maximum count of execution attempts
                      (>= 1 or ``None`` means infinity).
    :param exceptions: similar as exc but keyword only.
    :param giveup: is a predicate function which can decide by a given
    :param pause: is time gap between execution attempts.
    :param statistic_name: name filed for statistic instances
    """

    return asyncbackoff(
        attempt_timeout=None,
        deadline=None,
        exceptions=exceptions,
        giveup=giveup,
        max_tries=max_tries,
        pause=pause,
        statistic_class=RetryStatistic,
        statistic_name=statistic_name,
    )
